/* A program to demonstrate matrix multiplication
   using MPI.

   David Monismith
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <mpi.h>

int main(int argc, char ** argv)
{
  const int M_SIZE;     //The matrix size -- argv[1]

  int my_rank;
  int number_of_processes;
  int work_division;
  int mpi_error_code;
  int i, j, k;

  if(argc != 2)
  {
    printf("Error: no command line arguments were entered\n");
    exit(0);
  }

  /* Store the arguments as the appropriate constants */

  //Get the matrix size first -- we are using a square matrix, so we only need
  //one value.
  sscanf(argv[1], "%d", &M_SIZE);

  unsigned int iseed = (unsigned int)327;

  double * A = malloc(sizeof(double)*M_SIZE*M_SIZE);
  double * B = malloc(sizeof(double)*M_SIZE*M_SIZE);
  double ** C = malloc(sizeof(double *) * M_SIZE);

  for(i = 0; i < M_SIZE; i++)
    C[i] = malloc(sizeof(double)*M_SIZE);

  double * arrA;
  double * arrB;
  double * result;
  
  srand(iseed);

  mpi_error_code = MPI_Init(&argc, &argv);

    mpi_error_code = MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);
    mpi_error_code = MPI_Comm_size(MPI_COMM_WORLD, &number_of_processes);

    if(my_rank == 0)
    {
      double start = MPI_Wtime();
      printf("Entered process 0\n");
      /* Set up the arrays */
      for(i = 0; i < M_SIZE; i++)
      {
        for(j = 0; j < M_SIZE; j++)
        {
          A[i*M_SIZE + j] = (double)rand()/(double)RAND_MAX;
          B[i*M_SIZE + j] = (double)rand()/(double)RAND_MAX;
        }
      }

      /* Start computing values */

      work_division = M_SIZE*M_SIZE/(number_of_processes-1);

      printf("Work division is %d\n", work_division);


      int current_rank = 1;
      int unit_count = 0;

      for(i = current_rank; i < number_of_processes; i++)
      {
        mpi_error_code = MPI_Send(A, M_SIZE*M_SIZE, MPI_DOUBLE, i, 1,
                                  MPI_COMM_WORLD);
        mpi_error_code = MPI_Send(B, M_SIZE*M_SIZE, MPI_DOUBLE, i, 2,
                                  MPI_COMM_WORLD);
      }


      int count = 0;
      current_rank = 1;
      unit_count = 0;

      MPI_Request * reqArr = malloc(sizeof(MPI_Request)*
                                    (number_of_processes-1));
      double ** results = malloc(sizeof(double *) * (number_of_processes-1));
      int workToDo = 0;

      for(i = 0; i < number_of_processes-1; i++)
      {
        if(i < number_of_processes-2)
          workToDo = (M_SIZE*M_SIZE)/(number_of_processes - 1);
        else
          workToDo = M_SIZE*M_SIZE - (number_of_processes - 2)*
           ( M_SIZE*M_SIZE / (number_of_processes - 1) );
        printf("Recieving %d doubles from process %d\n", workToDo, i+1);
        results[i] = malloc(sizeof(double)*workToDo);
        mpi_error_code = MPI_Irecv(results[i], workToDo, MPI_DOUBLE, i+1, 0, 
          MPI_COMM_WORLD, &reqArr[i]);
      }

      int * index = malloc(sizeof(int)*(number_of_processes-1));
      MPI_Status * statuses = malloc(sizeof(MPI_Status)*
        (number_of_processes-1));
      int numComplete = 0;
      MPI_Waitsome(number_of_processes-1, reqArr, &numComplete, index, 
        statuses);
      int totalComplete = numComplete;
      printf("Total complete is %d\n", totalComplete);
      while(totalComplete < number_of_processes-1)
      {
        MPI_Waitsome(number_of_processes-1, reqArr, &numComplete, index, statuses);
        totalComplete += numComplete;
        printf("Total complete is %d\n", totalComplete);
      }
        
//      MPI_Waitall(M_SIZE*M_SIZE, reqArr, MPI_STATUSES_IGNORE);
      free((void *) reqArr);
      free((void *) index);
      free((void *) statuses);

      //Count over each of the sets of results and each element within
      //that set
      int resultsCounter = 0;
      int resultsProcess = 0;
      int numResults = (M_SIZE*M_SIZE)/(number_of_processes - 1);

      for( i = 0; i < M_SIZE; i++ )
      {
        for(j = 0; j < M_SIZE; j++)
        {
          C[i][j] = results[resultsProcess][resultsCounter];
          resultsCounter++;
          if(resultsProcess < number_of_processes-2 && resultsCounter >= numResults)
          {
            resultsCounter = 0;
            resultsProcess++;
          }
        }
      }

      for(i = 0; i < number_of_processes-1; i++)
        free((void *) results[i]);
      free(results);

      double endTime = MPI_Wtime() - start;
      printf("Completed matrix multiplication in %lf seconds.\n", endTime);
      

    }
    else
    {
      printf("Entered process %d\n", my_rank);
      int workToDo;

      if(my_rank < number_of_processes-1)
        workToDo = (M_SIZE*M_SIZE)/(number_of_processes - 1);
      else
        workToDo = M_SIZE*M_SIZE - (number_of_processes - 2)* 
         ( M_SIZE*M_SIZE / (number_of_processes - 1) );

      int standardWork = (M_SIZE*M_SIZE)/(number_of_processes - 1);
      int start = (my_rank-1)*standardWork;
      int end = start + workToDo;

      int d1_start = start / M_SIZE;
      int d2_start = start - d1_start*M_SIZE;
      int d1_end = end/M_SIZE;
      int d2_end = end - d1_end*M_SIZE;

      printf("d1_start = %d, d1_end = %d\n", d1_start, d1_end);
      printf("d2_start = %d, d2_end = %d\n", d2_start, d2_end);

      mpi_error_code = MPI_Recv(A, M_SIZE*M_SIZE, MPI_DOUBLE, 0, 1, 
                                MPI_COMM_WORLD, MPI_STATUS_IGNORE);
      mpi_error_code = MPI_Recv(B, M_SIZE*M_SIZE, MPI_DOUBLE, 0, 2, 
                                MPI_COMM_WORLD, MPI_STATUS_IGNORE);

      printf("Received matrices in process %d\n", my_rank);
      printf("Work to do is %d\n", workToDo);
      int workCount = 0;
      int innerStart = 0;
      int wcMultiplier = (my_rank-1)*(M_SIZE*M_SIZE/(number_of_processes-1));

      result = malloc(sizeof(double)*workToDo);

      for(i = d1_start; i < d1_end + 1 && workCount < workToDo; i++)
      {
        if(i == d1_start)
          innerStart = d2_start;
        else
          innerStart = 0;
        for(j = innerStart; j < M_SIZE && workCount < workToDo; j++)
        {
          result[workCount] = 0;
        
          for(k = 0; k < M_SIZE; k++)
          {
            result[workCount] += A[i*M_SIZE + k] * B[k*M_SIZE + j];
          }
          workCount++;
        }
      }
      mpi_error_code = MPI_Send(result, workToDo, MPI_DOUBLE, 
        0, 0, MPI_COMM_WORLD);
      free((void *) result);
      printf("Finished sending %d messages from process %d\n", workCount, my_rank);
    }
    free((void *) A);
    free((void *) B);

    for(i = 0; i < M_SIZE; i++)
      free((void *) C[i]);

    free((void *) C);
  mpi_error_code = MPI_Finalize();
}
